import os
import magic
import shutil
import hashlib
import logging
import requests
import subprocess
from bs4 import BeautifulSoup
from .base_handler import BaseHandler
from urllib.parse import urlparse, parse_qs
from ..common import PackageManager, temp_directory
from ..utils import download_file, temp_directory
from ..utils import extract_tar, extract_zip


class GemHandler(BaseHandler):
    def fetch(self):
        repo_url = self.construct_repo_url()
        self.repo_url = repo_url
        with temp_directory() as temp_dir:
            self.temp_dir = temp_dir
            if 'rubygem' in repo_url:
                self.fetch_file(repo_url)
                logging.info(f"File downloaded in {self.temp_dir}")
                self.unpack()
            else:
                self.clone_repo(repo_url)
                logging.info(f"Repo cloned to {self.temp_dir}")
            self.scan()

    def construct_repo_url(self):
        pkg_name = self.purl_details['name']
        pkg_version = self.purl_details['version']
        download_url = (
            f"https://rubygems.org/downloads/{pkg_name}-{pkg_version}.gem"
        )
        response = requests.get(download_url)
        if response.status_code == 200:
            return download_url
        else:
            api_url = f"https://rubygems.org/api/v1/gems/{pkg_name}.json"
            logging.info(f"api_url: {api_url}")
            response = requests.get(api_url)
            if response.status_code == 200:
                data = response.json()
                gem_url = data.get('gem_uri', '')
                if gem_url is None:
                    gem_url = ''
                source_code_url = data.get('source_code_uri', '')
                if source_code_url is None:
                    source_code_url = ''
                homepage_url = data.get('homepage_uri', '')
                if homepage_url is None:
                    homepage_url = ''
                repo_url = ''
                if 'rubygems.org' in gem_url:
                    return gem_url
                elif 'github.com' in source_code_url:
                    repo_url = source_code_url
                elif 'github.com' in homepage_url:
                    repo_url = homepage_url
                else:
                    logging.error(
                        f"Invalid source URL: {source_code_url} {homepage_url}"
                    )
                    exit()
            else:
                logging.error(f"Failed to fetch data: {response.status_code}")
                exit()
            return f"{repo_url}.git"

    def unpack(self):
        if self.temp_dir:
            package_file_path = os.path.join(
                self.temp_dir,
                "downloaded_file"
            )
            mime = magic.Magic(mime=True)
            mimetype = mime.from_file(package_file_path)
            if 'gzip' in mimetype:
                extract_zip(package_file_path, self.temp_dir)
                logging.info(f"Unpacked package in {self.temp_dir}")
            elif 'tar' in mimetype:
                extract_tar(package_file_path, self.temp_dir)
                extract_tar(self.temp_dir+'/data.tar.gz', self.temp_dir)
                # extract_tar(self.temp_dir+'/metadata.gz', self.temp_dir)
                logging.info(f"Unpacked package in {self.temp_dir}")
            else:
                logging.error(f"MimeType not supported {mimetype}")
                logging.error(f"Error unpacking file in {self.temp_dir}")
                exit()

    def scan(self):
        results = {}
        logging.info("Scanning package contents...")
        files = PackageManager.scan_for_files(
            self.temp_dir, ['COPYRIGHT', 'NOTICES', 'LICENSE', 'COPYING']
        )
        results['license_files'] = files
        copyhits = PackageManager.scan_for_copyright(self.temp_dir)
        results['copyrights'] = copyhits
        pkg_name = self.purl_details['name']
        results['license'] = self.get_license(pkg_name)
        results['url'] = self.repo_url
        self.results = results

    def generate_report(self):
        logging.info("Generating report based on the scanned data...")
        return self.results

    def get_license(self, pkg_name):
        url = f"https://rubygems.org/api/v1/gems/{pkg_name}.json"
        response = requests.get(url)
        if response.status_code == 200:
            data = response.json()
            license_info = data.get('licenses') or data.get('license')
            if license_info:
                if isinstance(license_info, list):
                    return ', '.join(license_info)
                return license_info
            else:
                return ''
        else:
            logging.error("Can't obtain data from Crates.IO")
            return ''

    def fetch_file(self, url):
        response = requests.get(url)
        if response.status_code == 200:
            file_data = response.content
            package_file_path = os.path.join(
                self.temp_dir,
                "downloaded_file"
            )
            with open(package_file_path, "wb") as file:
                file.write(file_data)
            logging.info("File downloaded successfully.")
        else:
            raise ConnectionError("Failed to download the file.")

    def clone_repo(self, repo_url):
        repo = repo_url[0]
        try:
            subprocess.run(
                ["git", "clone", repo, self.temp_dir],
                check=True,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL
            )
            # Here should clone base on version. Check GoLang
            logging.info(f"Repository cloned successfully to {self.temp_dir}")
        except subprocess.CalledProcessError as e:
            print(f"Failed to clone repository: {e}")
            # shutil.rmtree(self.temp_dir)
            raise



import os
import re
import json
import magic
import pickle
import logging
from pkg_resources import resource_filename
from oslili import LicenseAndCopyrightIdentifier
from urllib.parse import unquote, urlparse, parse_qs
from urllib.parse import urlparse, parse_qs, unquote
from .utils import download_file, temp_directory, extract_zip, extract_tar


class PackageManager:
    @staticmethod
    def parse_purl(purl):
        logging.info(f"parse_purl {purl}")
        result = {}
        url = urlparse(purl)
        if url.scheme != 'pkg':
            raise ValueError("Invalid PURL scheme")
        path_parts = url.path.strip('/').split('/')
        if len(path_parts) < 2:
            raise ValueError(
                "Invalid PURL format. Expected at least pkg:type/name@version"
            )
        result['type'] = path_parts[0]
        name_with_version = path_parts[-1]
        if len(path_parts) >= 3:
            result['namespace'] = unquote(path_parts[1])
        else:
            result['namespace'] = None
        name_version = name_with_version.split('@', 1)
        result['name'] = unquote(name_version[0])
        if len(name_version) > 1:
            result['version'] = name_version[1]
        else:
            result['version'] = None
        # Small ugly hack for Go
        if 'golang' in result['type']:
            if len(path_parts) == 5:
                result['name'] = path_parts[3]
                result['version'] = None
            if len(path_parts) <= 2:
                print('purl:', purl)
                raise ValueError(
                    "Invalid PURL format."
                )
        result['qualifiers'] = parse_qs(url.query)
        result['subpath'] = unquote(url.fragment) if url.fragment else None
        result['fullparts'] = path_parts
        return result

    @staticmethod
    def unpack_package(file_path):
        with temp_directory() as temp_dir:
            if file_path.endswith('.zip'):
                extract_zip(file_path, temp_dir)
            elif file_path.endswith('.tar.gz') or file_path.endswith('.tar'):
                extract_tar(file_path, temp_dir)
            else:
                print(f"Unsupported file format for {file_path}")
                raise ValueError("Unsupported file format")
            print(f"Package unpacked to {temp_dir}")
            return temp_dir

    @staticmethod
    def scan_for_files(temp_dir, patterns):
        identifier = LicenseAndCopyrightIdentifier()
        found_files = []
        for root, dirs, files in os.walk(temp_dir):
            for file in files:
                if any(
                    re.match(pattern, file, re.IGNORECASE)
                    for pattern in patterns
                ):
                    file_path = os.path.join(root, file)
                    file_text = PackageManager.read_file_content(file_path)
                    if file_text:
                        spdx_code, proba = identifier.identify_license(
                            file_text
                        )
                        found_files.append({
                            "file": file_path,
                            "content": file_text,
                            "spdx": spdx_code,
                            "oslili": proba,
                        })
        return found_files

    @staticmethod
    def is_readable_text_file(file_path):
        mime = magic.Magic(mime=True)
        mimetype = mime.from_file(file_path)
        if mimetype:
            if mimetype.startswith('text'):
                return True
            if mimetype == 'application/octet-stream':
                return True
        return False

    @staticmethod
    def read_file_content(file_path, encodings=['utf-8', 'iso-8859-1']):
        if PackageManager.is_readable_text_file(file_path):
            for encoding in encodings:
                try:
                    with open(file_path, 'r', encoding=encoding) as file:
                        return file.read()
                except UnicodeDecodeError:
                    continue
            return None
        else:
            logging.info(f"read_file_content failed {file_path}")
            return None

    @staticmethod
    def scan_for_copyright(temp_dir):
        naive_bayes_model_path = resource_filename(
            __name__,
            'datasets/naive_bayes_model.pkl'
        )
        tfidf_data_path = resource_filename(
            __name__,
            'datasets/tfidf_data.pkl'
        )
        with open(naive_bayes_model_path, 'rb') as f:
            classifier = pickle.load(f)
        with open(tfidf_data_path, 'rb') as f:
            _, _, _, _, vectorizer = pickle.load(f)
        copyrights = []
        pattern = r"[^0-9<>,.()@a-zA-Z-\s]+"
        for root, _, files in os.walk(temp_dir):
            for file in files:
                file_path = os.path.join(root, file)
                if PackageManager.is_readable_text_file(file_path):
                    try:
                        with open(file_path, 'r', encoding='utf-8') as f:
                            for line in f:
                                clean_line = line.strip().lower()
                                if (
                                    "copyright " in clean_line
                                    and len(clean_line) <= 50
                                    and "yyyy" not in clean_line
                                ):
                                    clean_line = re.sub(
                                        pattern, "", clean_line
                                    )
                                    if (
                                        clean_line.startswith('copyright')
                                        or " copyright" in clean_line
                                    ):
                                        input_tfidf = vectorizer.transform(
                                            [clean_line]
                                        )
                                        prediction = classifier.predict(
                                            input_tfidf
                                        )[0]
                                        if 'copyright' in prediction:
                                            copyrights.append({
                                                "file": file_path,
                                                "line": clean_line
                                            })
                    except UnicodeDecodeError:
                        continue
        return copyrights

    @staticmethod
    def serialize_output(data):
        return json.dumps(data, indent=4)
